from __future__ import absolute_import

from ._doc_parse_tools import (merge_google_napoleon_docs, merge_numpy_docs,
                               merge_numpy_napoleon_docs, merge_rest_docs)

""" Docstring inheritance-style implementations.

    To implement your own inheritance file, simply write a function that fits the template:

            def your_style(prnt_doc, child_doc):
                ''' Merges parent and child docstrings

                    Parameters
                    ----------
                    prnt_cls_doc: Optional[str]
                    child_doc: Optional[str]

                    Returns
                    ------
                    Optional[str]
                        The merged docstring that will be utilized.'''
                return final_docstring

    and log this using `custom_inherit.add_style(your_style)`. To permanently save your function,
    define your function within custom_inherit/_style_store.py, and log it in custom_inherit.style_store.__all__.
    Your style will then be available as 'your_style' (i.e. whatever you named the function).
"""

# All built-in styles must be logged in the __all__ field.
__all__ = [
    "parent",
    "numpy",
    "reST",
    "google",
    "numpy_napoleon",
    "google_with_merge",
    "numpy_napoleon_with_merge",
    "numpy_with_merge"
]


def parent(prnt_doc, child_doc):
    """ Wherever the docstring for a child-class' attribute (or for the class itself) is
    `None`, inherit the corresponding docstring from the parent.

    *NOTE* As of Python 3.5, this is the default behavior of the built-in function
    inspect.getdoc, and thus this style is deprecated Python 3.(>=5). """
    return child_doc if child_doc is not None else prnt_doc


def numpy(prnt_doc, child_doc):
    """ Merges numpy-styled docstrings from the parent and child.

    Specifically, any docstring section that appears in the parent's docstring that
    is not present in the child's is inherited. Otherwise, the child's docstring
    section is utilized. An exception to this is if the parent docstring contains a
    "Raises" section, but the child's attribute's docstring contains a "Returns" or
    "Yields" section instead. In this instance, the "Raises" section will not appear in the
    inherited docstring.

    Any whitespace that can be uniformly removed from a docstring's second line and onwards is
    removed. Sections in the resulting docstring will be separated by a single blank line.

    For details on the numpy docstring style, see:
    https://github.com/numpy/numpy/blob/master/doc/HOWTO_DOCUMENT.rst.txt

    Example:
        - parent's docstring:

            ''' Parent's line

                Parameters
                ----------
                x: int
                    description of x
                y: Union[None, int]
                    description of y

                Raises
                ------
                NotImplemented Error'''

        - child's docstring:

            ''' Child's line

                Returns
                -------
                int

                Notes
                -----
                notes blah blah'''

        - docstring that is ultimately inherited:

            ''' Child's line

                Parameters
                ----------
                x: int
                    description of x
                y: Union[None, int]
                    description of y

                Returns
                -------
                int

                Notes
                -----
                notes blah blah'''"""
    return merge_numpy_docs(prnt_doc, child_doc)


def reST(prnt_doc, child_doc):
    """ Merge two reST-style docstrings into a single docstring.

    Given the reST-style docstrings from a parent and child's attributes, merge the docstring
    sections such that the child's section is used, wherever present, otherwise the parent's
    section is used.

    Sections are delimited by any type of reST section title. For more details, see:
    http://docutils.sourceforge.net/docs/ref/rst/restructuredtext.html#sections

    Any whitespace that can be uniformly removed from a docstring's second line and onwards is
    removed. Sections in the resulting docstring will be separated by a single blank line.

    Example:
      parent - ''' Header1
                   -------
                   parent's content for Header 1
                       - indented material
                   Header2
                   ~~~~~~~
                   content for Header2'''

      child -  ''' Front-matter
                   ~~~~~~~~~
                   NewHeader
                   ~~~~~~~~~
                   content for NewHeader

                   Header2
                   +++++++
                   child's content for Header2'''

      merged - ''' Front-matter

                   Header1
                   -------
                   content for Header 1
                       - indented material

                   Header2
                   +++++++
                   child's content for Header2

                   ~~~~~~~~~
                   NewHeader
                   ~~~~~~~~~
                   content for NewHeader '''
            """
    return merge_rest_docs(prnt_doc, child_doc)


def numpy_napoleon(prnt_doc, child_doc):
    """ Behaves identically to the 'numpy' style, but abides by the docstring sections
    specified by the "Napoleon" standard.

    For more info regarding the Napoleon standard, see:
    http://sphinxcontrib-napoleon.readthedocs.io/en/latest/index.html#docstring-sections

    Example:
        - parent's docstring:

            ''' Parent's line

                Keyword Arguments
                -----------------
                x: int
                    description of x
                y: Union[None, int]
                    description of y

                Raises
                ------
                NotImplemented Error'''

        - child's docstring:

            ''' Child's line

                Returns
                -------
                int

                Notes
                -----
                notes blah blah'''

        - docstring that is ultimately inherited:

            ''' Child's line

                Keyword Arguments
                -----------------
                x: int
                    description of x
                y: Union[None, int]
                    description of y

                Returns
                -------
                int

                Notes
                -----
                notes blah blah'''
    """
    return merge_numpy_napoleon_docs(prnt_doc, child_doc)


def google(prnt_doc, child_doc):
    """ Merges google-styled docstrings from the parent and child, abiding to the docstring sections
    specified by the "Napoleon" standard.

    Specifically, any docstring section that appears in the parent's docstring that
    is not present in the child's is inherited. Otherwise, the child's docstring
    section is utilized. An exception to this is if the parent docstring contains a
    "Raises" section, but the child's attribute's docstring contains a "Returns" or
    "Yields" section instead. In this instance, the "Raises" section will not appear in the
    inherited docstring.

    Any whitespace that can be uniformly removed from a docstring's second line and onwards is
    removed. Sections in the resulting docstring will be separated by a single blank line.

    For more info regarding the Napoleon standard, see:
    http://sphinxcontrib-napoleon.readthedocs.io/en/latest/index.html#docstring-sections

    For details on the google docstring style, see:
    http://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_google.html#example-google

    Example:
        - parent's docstring:

            ''' Parent's line

                Args:
                    x: int
                        description of x
                    y: Union[None, int]
                        description of y

                Raises:
                    NotImplemented Error'''

        - child's docstring:

            ''' Child's line

                Returns:
                    int

                Notes:
                    notes blah blah'''

        - docstring that is ultimately inherited:

            ''' Child's line

                Parameters:
                    x: int
                        description of x
                    y: Union[None, int]
                        description of y

                Returns:
                    int

                Notes:
                    notes blah blah'''
        """
    return merge_google_napoleon_docs(prnt_doc, child_doc)


def google_with_merge(prnt_doc, child_doc):
    """ Behaves identically to the 'google' style, but also merges sections that
    overlap, instead of only keeping the child's section. All sections are
    concerned except sections "Short Summary", "Example" and "Examples" (or
    coresponding aliases) for which the 'google' style applies.

    Example: - parent's docstring:

            ''' Parent's line

                Args:
                    x: int
                        description of x
                    y: Union[None, int]
                        description of y

                Raises:
                    NotImplemented Error

                Example:
                    >>> parent_func(x=3, y=None)
                    NotImplementedError:'''

        - child's docstring:

            ''' Child's line

                Args:
                    z: Union[None, int]
                        description of z

                Returns:
                    int

                Example:
                    >>> child_func(x=3, y=None, z=4)
                    7

                Notes:
                    notes blah blah'''

        - docstring that is ultimately inherited:

            ''' Child's line

                Parameters:
                    x: int
                        description of x
                    y: Union[None, int]
                        description of y
                    z: Union[None, int]
                        description of z

                Returns:
                    int

                Example:
                    >>> child_func(x=3, y=None, z=4)
                    7

                Notes:
                    notes blah blah'''
    """
    return merge_google_napoleon_docs(prnt_doc, child_doc, merge_within_sections=True)


def numpy_napoleon_with_merge(prnt_doc, child_doc):
    """
    Behaves identically to the 'numpy_napoleon' style, but also merges sections
    that overlap, instead of only keeping the child's section. All sections are
    concerned except sections "Short Summary", "Example" and "Examples" (or
    coresponding aliases) for which the 'numpy_napoleon' style behaviour
    applies.

    Example: - parent's docstring:

            ''' Parent's line

                Keyword Arguments
                -----------------
                x: int
                    description of x
                y: Union[None, int]
                    description of y

                Raises
                ------
                NotImplemented Error

                Example
                -------
                >>> parent_func(x=3, y=None)
                NotImplementedError:'''

        - child's docstring:

            ''' Child's line

                Keyword Arguments
                -----------------
                z: Union[None, int]
                    description of z

                Returns
                -------
                int

                Notes
                -----
                notes blah blah

                Example
                -------
                >>> child_func(x=3, y=None, z=4)
                7'''

        - docstring that is ultimately inherited:

            ''' Child's line

                Keyword Arguments
                -----------------
                x: int
                    description of x
                y: Union[None, int]
                    description of y
                z: Union[None, int]
                    description of z

                Returns
                -------
                int

                Notes
                -----
                notes blah blah

                Example
                -------
                >>> child_func(x=3, y=None, z=4)
                7'''
    """
    return merge_numpy_napoleon_docs(prnt_doc, child_doc, merge_within_sections=True)


def numpy_with_merge(prnt_doc, child_doc):
    """
    Behaves identically to the 'numpy' style, but also merges sections that
    overlap, instead of only keeping the child's section. All sections are
    concerned except sections "Short Summary", "Extended Summary", "Deprecation
    Warning" and "Examples" for which the 'numpy' style behaviour applies.

    Example:
        - parent's docstring:

            ''' Parent's line

                Parameters
                ----------
                x: int
                    description of x
                y: Union[None, int]
                    description of y

                Raises
                ------
                NotImplemented Error

                Example
                -------
                >>> parent_func(x=3, y=None)
                NotImplementedError:'''

        - child's docstring:

            ''' Child's line

                Parameters
                ----------
                z: Union[None, int]
                    description of z

                Returns
                -------
                int

                Notes
                -----
                notes blah blah

                Example
                -------
                >>> child_func(x=3, y=None, z=4)
                7'''

        - docstring that is ultimately inherited:

            ''' Child's line

                Parameters
                ----------
                x: int
                    description of x
                y: Union[None, int]
                    description of y
                z: Union[None, int]
                    description of z

                Returns
                -------
                int

                Notes
                -----
                notes blah blah

                Example
                -------
                >>> child_func(x=3, y=None, z=4)
                7'''
    """
    return merge_numpy_docs(prnt_doc, child_doc, merge_within_sections=True)



from __future__ import absolute_import

from collections import OrderedDict
from inspect import cleandoc

__all__ = ["merge_google_napoleon_docs", "merge_numpy_napoleon_docs"]


def parse_napoleon_doc(doc, style):
    """ Extract the text from the various sections of a numpy-formatted docstring.

    Parameters
    ----------
    doc: Union[str, None]
        The docstring to parse.

    style: str
        'google' or 'numpy'

    Returns
    -------
    OrderedDict[str, Union[None,str]]
        The extracted numpy-styled docstring sections."""

    napoleon_sections = [
        "Short Summary",
        "Attributes",
        "Methods",
        "Warning",
        "Note",
        "Parameters",
        "Other Parameters",
        "Keyword Arguments",
        "Returns",
        "Yields",
        "Raises",
        "Warns",
        "See Also",
        "References",
        "Todo",
        "Example",
        "Examples",
    ]

    aliases = {
        "Args": "Parameters",
        "Arguments": "Parameters",
        "Keyword Args": "Keyword Arguments",
        "Return": "Returns",
        "Warnings": "Warning",
        "Yield": "Yields",
    }

    doc_sections = OrderedDict([(key, None) for key in napoleon_sections])

    if not doc:
        return doc_sections

    assert style in ("google", "numpy")

    doc = cleandoc(doc)
    lines = iter(doc.splitlines())

    key = "Short Summary"
    body = []
    while True:
        try:
            line = next(lines).rstrip()
            header = (
                line
                if style == "numpy"
                else (line[:-1] if line.endswith(":") else line)
            )
            if header and (header in doc_sections or header in aliases):
                doc_sections[aliases.get(key, key)] = (
                    "\n".join(body).rstrip() if body else None
                )
                body = []
                key = header
                if style == "numpy":
                    next(lines)  # skip section delimiter
            else:
                body.append(line)
        except StopIteration:
            doc_sections[aliases.get(key, key)] = "\n".join(body)
            break
    return doc_sections


def merge_section(key, prnt_sec, child_sec, style, merge_within_sections=False):
    """ Synthesize a output napoleon docstring section.

    Parameters
    ----------
    key: str
        The napoleon-section being merged.
    prnt_sec: Optional[str]
        The docstring section from the parent's attribute.
    child_sec: Optional[str]
        The docstring section from the child's attribute.
    Returns
    -------
    Optional[str]
        The output docstring section."""

    napoleon_sections_that_cant_merge = [
        "Short Summary",
        "Example",
        "Examples",
    ]

    if prnt_sec is None and child_sec is None:
        return None

    assert style in ("google", "numpy")

    if key == "Short Summary":
        header = ""
    else:
        if style == "numpy":
            header = "\n".join((key, "".join("-" for i in range(len(key))), ""))
        else:
            header = "\n".join((key + ":", ""))
    if merge_within_sections and key not in napoleon_sections_that_cant_merge:
        if child_sec is None:
            body = prnt_sec
        elif prnt_sec is None:
            body = child_sec
        else:
            body = '\n'.join((prnt_sec, child_sec))
    else:
        body = prnt_sec if child_sec is None else child_sec

    return header + body


def merge_all_sections(prnt_sctns, child_sctns, style, merge_within_sections=False):
    """ Merge the doc-sections of the parent's and child's attribute into a single docstring.

    Parameters
    ----------
    prnt_sctns: OrderedDict[str, Union[None,str]]
    child_sctns: OrderedDict[str, Union[None,str]]

    Returns
    -------
    str
        Output docstring of the merged docstrings."""
    doc = []

    prnt_only_raises = prnt_sctns["Raises"] and not (
        prnt_sctns["Returns"] or prnt_sctns["Yields"]
    )
    if prnt_only_raises and (child_sctns["Returns"] or child_sctns["Yields"]):
        prnt_sctns["Raises"] = None

    for key in prnt_sctns:
        sect = merge_section(
            key,
            prnt_sctns[key],
            child_sctns[key],
            style,
            merge_within_sections=merge_within_sections
        )
        if sect is not None:
            doc.append(sect)
    return "\n\n".join(doc) if doc else None


def merge_numpy_napoleon_docs(prnt_doc=None, child_doc=None, merge_within_sections=False):
    """ Merge two numpy-style docstrings into a single docstring, according to napoleon docstring sections.

    Given the numpy-style docstrings from a parent and child's attributes, merge the docstring
    sections such that the child's section is used, wherever present, otherwise the parent's
    section is used.

    Any whitespace that can be uniformly removed from a docstring's second line and onwards is
    removed. Sections will be separated by a single blank line.

    Aliased docstring sections are normalized. E.g Args, Arguments -> Parameters

    Parameters
    ----------
    prnt_doc: Optional[str]
        The docstring from the parent.
    child_doc: Optional[str]
        The docstring from the child.

    Returns
    -------
    Union[str, None]
        The merged docstring. """
    style = "numpy"
    return merge_all_sections(
        parse_napoleon_doc(prnt_doc, style),
        parse_napoleon_doc(child_doc, style),
        style,
        merge_within_sections=merge_within_sections
    )


def merge_google_napoleon_docs(prnt_doc=None, child_doc=None, merge_within_sections=False):
    """ Merge two google-style docstrings into a single docstring, according to napoleon docstring sections.

        Given the google-style docstrings from a parent and child's attributes, merge the docstring
    sections such that the child's section is used, wherever present, otherwise the parent's
    section is used.

    Any whitespace that can be uniformly removed from a docstring's second line and onwards is
    removed. Sections will be separated by a single blank line.

    Aliased docstring sections are normalized. E.g Args, Arguments -> Parameters

    Parameters
    ----------
    prnt_doc: Optional[str]
        The docstring from the parent.
    child_doc: Optional[str]
        The docstring from the child.

    Returns
    -------
    Union[str, None]
        The merged docstring. """
    style = "google"
    return merge_all_sections(
        parse_napoleon_doc(prnt_doc, style),
        parse_napoleon_doc(child_doc, style),
        style,
        merge_within_sections=merge_within_sections
    )



try:
    basestring
except NameError:
    basestring = str  # Python 2 -> 3 alias

""" Exposes decorator class."""

__all__ = ["DocInheritDecorator"]


class DocInheritDecorator(object):
    """ A decorator that merges provided parent docstring with the docstring of the decorated
    function/method/property.

    Methods
    -------
    doc_merger(prnt_attr_doc, child_doc)
        Merges the parent and child docstrings into a single docstring.

    Notes
    -----
    When utilized as the inner-most decorator, this decorator can be used on functions decorated functions,
    methods, properties, static methods, class methods, and their abstract counterparts."""

    def __init__(self, prnt_doc):
        """
        Parameters
        ----------
        prnt_doc : Union[str, Any]
            The docstring, or object of which the docstring is utilized as the
            parent docstring during the docstring merge."""
        self.prnt_doc = (
            prnt_doc if isinstance(prnt_doc, basestring) else prnt_doc.__doc__
        )

    def __call__(self, func):
        """
        Parameters
        ----------
        func : FunctionType
            The function/method/property to be decorated

        Returns
        -------
        FunctionType
            The decorated function/method/property whose docstring is given by
            DocInheritDecorator.doc_merger(prnt_attr_doc, child_doc)"""
        func.__doc__ = self.doc_merger(self.prnt_doc, func.__doc__)
        return func

    @staticmethod
    def doc_merger(prnt_attr_doc, child_doc):
        """ Merges the parent and child docstrings into a single docstring.

        Parameters
        ----------
        prnt_attr_doc: Union[None, str]
        child_doc: Union[None, str]

        Raises
        ------
        NotImplementedError"""
        raise NotImplementedError



from __future__ import absolute_import

from .napoleon_parse_tools import (merge_google_napoleon_docs,
                                   merge_numpy_napoleon_docs)
from .numpy_parse_tools import merge_numpy_docs
from .rest_parse_tools import merge_rest_docs

__all__ = [
    "merge_numpy_docs",
    "merge_rest_docs",
    "merge_numpy_napoleon_docs",
    "merge_google_napoleon_docs",
]



from __future__ import absolute_import

from abc import abstractproperty
from types import FunctionType, MethodType

""" Exposes abstract base meta class to be inherited by inheritance-style meta classes.

    This metaclass merges the respective docstrings of a parent class and its child, and their
    properties, methods (including classmethod, staticmethod, decorated methods)

    This merge-style must be implemented via the static methods `class_doc_inherit`
    and `attr_doc_inherit`. See custom_inherit/_style_store.py for such implementations."""

__all__ = ["DocInheritorBase"]


class DocInheritorBase(type):
    """ A metaclass that merges the respective docstrings of a parent class and of its child, along with their
    properties, methods (including classmethod, staticmethod, decorated methods).

    This merge-style must be implemented via the static methods `class_doc_inherit`
    and `attr_doc_inherit`, which are set within `custom_inherit.DocInheritMeta`."""

    include_special_methods = False

    def __new__(mcs, class_name, class_bases, class_dict):
        # inherit class docstring: the docstring is constructed by traversing
        # the mro for the class and merging their docstrings, with each next
        # docstring as serving as the 'parent', and the accumulated docstring
        # serving as the 'child'
        this_doc = class_dict.get("__doc__", None)
        for mro_cls in (mro_cls for base in class_bases for mro_cls in base.mro()):
            prnt_cls_doc = mro_cls.__doc__
            if prnt_cls_doc is not None:
                if prnt_cls_doc == "The most base type":
                    prnt_cls_doc = None
            this_doc = mcs.class_doc_inherit(prnt_cls_doc, this_doc)

        class_dict["__doc__"] = this_doc

        # inherit docstring for method, static-method, class-method, abstract-method, decorated-method, and property
        for attr, attribute in class_dict.items():
            is_doc_type = isinstance(
                attribute,
                (FunctionType, MethodType, classmethod, staticmethod, property),
            )
            if (
                (attr.startswith("__") and attr.endswith("__") and not mcs.include_special_methods) or
                not is_doc_type
            ):
                continue

            is_static_or_class = isinstance(attribute, (staticmethod, classmethod))
            child_attr = attribute if not is_static_or_class else attribute.__func__

            prnt_attr_doc = None
            for mro_cls in (
                mro_cls
                for base in class_bases
                for mro_cls in base.mro()
                if hasattr(mro_cls, attr)
            ):
                prnt_attr_doc = getattr(mro_cls, attr).__doc__

                if prnt_attr_doc is not None:
                    break

            if prnt_attr_doc is None:
                continue

            doc = mcs.attr_doc_inherit(prnt_attr_doc, child_attr.__doc__)
            try:
                child_attr.__doc__ = doc
            # property.__doc__ is read-only in Python 2 (TypeError), 3.3 - 3.4 (AttributeError)
            except (TypeError, AttributeError) as err:
                if type(child_attr) in (property, abstractproperty):
                    new_prop = property(
                        fget=child_attr.fget,
                        fset=child_attr.fset,
                        fdel=child_attr.fdel,
                        doc=doc,
                    )
                    if isinstance(child_attr, abstractproperty):
                        new_prop = abstractproperty(new_prop)
                    class_dict[attr] = new_prop
                else:
                    raise type(err)(err)

        return type.__new__(mcs, class_name, class_bases, class_dict)

    @staticmethod
    def class_doc_inherit(prnt_cls_doc, child_doc):
        """ Merge the docstrings of a parent class and its child.

        Parameters
        ----------
        prnt_cls_doc: Union[None, str]
        child_doc: Union[None, str]

        Raises
        ------
        NotImplementedError"""
        raise NotImplementedError

    @staticmethod
    def attr_doc_inherit(prnt_attr_doc, child_doc):
        """ Merge the docstrings of method or property from parent class and the corresponding
        attribute of its child.

        Parameters
        ----------
        prnt_attr_doc: Union[None, str]
        child_doc: Union[None, str]

        Raises
        ------
        NotImplementedError

        Notes
        -----
        This works for properties, methods, static methods, class methods, and
        decorated methods/properties."""
        raise NotImplementedError



from __future__ import absolute_import
from distutils.core import setup
from setuptools import find_packages
import versioneer

setup(
    name="custom_inherit",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    package_dir={"": "src"},
    packages=find_packages(where="src", exclude=["tests", "tests.*"]),
    description="A Python package that provides customized docstring inheritance\
       schemes between derived classes and their parents.",
    author="@rsokl",
    author_email="rsoklaski@gmail.com",
    url="https://github.com/rsokl/custom_inherit",
    download_url="https://github.com/rsokl/custom_inherit/tarball/" + versioneer.get_version(),
    license="MIT",
)



from __future__ import absolute_import as _absolute_import

from abc import ABCMeta as _ABCMeta

from ._decorator_base import DocInheritDecorator as _DocInheritDecorator
from ._metaclass_base import DocInheritorBase as _DocInheritorBase
from . import _style_store
from ._style_store import (
    google, numpy, numpy_napoleon, parent, reST,
    google_with_merge, numpy_napoleon_with_merge, numpy_with_merge
)
from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

try:
    _basestring = basestring
except NameError:
    _basestring = str  # Python 2 -> 3 alias


__all__ = ["DocInheritMeta", "doc_inherit", "store", "add_style", "remove_style"]


def _check_style_function(style_func):
    out = style_func("", "")
    if not isinstance(out, _basestring) and out is not None:
        raise TypeError
    return None


class _Store(object):
    """ A dictionary-like object that stores the styles available for the doc-inheritance metaclass and decorator,
   respectively.

   Only callable objects with the signature: f(Optional[str], Optional[str]) -> Optional[str]
   can be stored. If f is a valid callable, then _Store()[f] -> f."""

    def __init__(self, *args, **kwargs):
        self._store = dict()
        self.update(*args, **kwargs)

    def __repr__(self):
        return repr(self._store)

    def __str__(self):
        out_str = "The available stored styles are: "
        styles = "\n".join("\t- " + style for style in sorted(self.keys()))
        return "\n".join((out_str, styles))

    def __setitem__(self, style_name, style_func):
        """ Make available a new function for merging a 'parent' and 'child' docstring.

        Parameters
        ----------
        style_name : Any
            The identifier of the style being logged
        style_func: Callable[[Optional[str], Optional[str]], Optional[str]]
            The style function that merges two docstrings into a single docstring."""
        try:
            _check_style_function(style_func)
        except TypeError:
            raise TypeError(
                "The style store only stores callables of the form: "
                "\n\tstyle_func(Optional[str], Optional[str]) -> Optional[str]"
            )
        self._store[style_name] = style_func

    def __getitem__(self, item):
        """ Given a valid style-ID, retrieve a stored style. If a valid function (callable) is
        supplied, return it in place.

        Parameters
        ----------
        item : Union[Any, Callable[Optional[str], Optional[str]], Optional[str]]
            A valid style-ID or style-function."""
        try:
            return self._store[item]
        except KeyError:
            try:
                _check_style_function(item)
                return item
            except (TypeError, ValueError):
                raise TypeError(
                    "Either a valid style name or style-function must be specified"
                )

    def keys(self):
        """  D.keys() -> a set-like object providing a view on D's keys"""
        return self._store.keys()

    def pop(self, *args):
        """ D.pop(k[,d]) -> v, remove specified key and return the corresponding value.
        If key is not found, d is returned if given, otherwise KeyError is raised. """
        if len(args) < 3:
            return self._store.pop(*args)
        else:
            raise TypeError(
                "pop expected at most 2 arguments, got {}".format(len(args))
            )

    def update(self, *args, **kwargs):
        """ D.update([E, ]**F) -> None.  Update D from dict/iterable E and F.
        If E is present and has a .keys() method, then does:  for k in E: D[k] = E[k]"""
        if len(args) > 1:
            raise TypeError("update expected at most 1 arguments, got %d" % len(args))

        for key, value in dict(*args, **kwargs).items():
            self[key] = value

    def values(self):
        """ D.values() -> an object providing a view on D's values."""
        return self._store.values()

    def items(self):
        """ D.items() -> a set-like object providing a view on D's items"""
        return self._store.items()

store = _Store([(key, getattr(_style_store, key)) for key in _style_store.__all__])


def add_style(style_name, style_func):
    """ Make available a new function for merging a 'parent' and 'child' docstring.

    Parameters
    ----------
    style_name : Any
        The identifier of the style being logged
    style_func: Callable[[Optional[str], Optional[str]], Optional[str]]
        The style function that merges two docstrings into a single docstring."""
    store[style_name] = style_func


def remove_style(style):
    """ Remove the specified style from the style store.

    Parameters
    ----------
    style: Any
        The inheritance-scheme style ID to be removed."""
    if style in store:
        store.pop(style)


def DocInheritMeta(style="parent", abstract_base_class=False, include_special_methods=False):
    """ A metaclass that merges the respective docstrings of a parent class and of its child, along with their
    properties, methods (including classmethod, staticmethod, decorated methods).

    Parameters
    ----------
    style: Union[Any, Callable[[str, str], str]], optional (default: "parent")
        A valid inheritance-scheme style ID or function that merges two docstrings.

    abstract_base_class: bool, optional(default: False)
        If True, the returned metaclass inherits from abc.ABCMeta.

        Thus a class that derives from DocInheritMeta(style="numpy", abstract_base_class=True)
        will be an abstract base class, whose derived classes will inherit docstrings
        using the numpy-style inheritance scheme.

    include_special_methods: bool, optional (defaul: False)
        Wether special methods of class (i.e. starting en ending with "__") are included in the docstring
        inheritance process.


    Returns
    -------
    custom_inherit.DocInheritorBase"""

    merge_func = store[style]
    metaclass = _DocInheritorBase
    metaclass.include_special_methods = include_special_methods
    metaclass.class_doc_inherit = staticmethod(merge_func)
    metaclass.attr_doc_inherit = staticmethod(merge_func)

    return (
        metaclass
        if not abstract_base_class
        else type("abc" + metaclass.__name__, (_ABCMeta, metaclass), {})
    )


def doc_inherit(parent, style="parent"):
    """ Returns a function/method decorator that, given `parent`, updates the docstring of the decorated
    function/method based on the specified style and the corresponding attribute of `parent`.

    Parameters
    ----------
    parent : Union[str, Any]
        The docstring, or object of which the docstring is utilized as the
        parent docstring during the docstring merge.

    style : Union[Any, Callable[[str, str], str]], optional (default: "parent")
        A valid inheritance-scheme style ID or function that merges two docstrings.

    Returns
    -------
    custom_inherit.DocInheritDecorator

    Notes
    -----
    `doc_inherit` should always be used as the inner-most decorator when being used in
    conjunction with other decorators, such as `@property`, `@staticmethod`, etc."""

    merge_func = store[style]
    decorator = _DocInheritDecorator
    decorator.doc_merger = staticmethod(merge_func)
    return decorator(parent)



from __future__ import absolute_import

from collections import OrderedDict
from inspect import cleandoc
from string import punctuation

__all__ = ["merge_rest_docs"]


def is_delimiter(line):
    """ True if a line consists only of a single punctuation character."""
    return bool(line) and line[0] in punctuation and line[0] * len(line) == line


def parse_rest_doc(doc):
    """ Extract the headers, delimiters, and text from reST-formatted docstrings.

    Parameters
    ----------
    doc: Union[str, None]

    Returns
    -------
    Dict[str, Section] """

    class Section(object):
        def __init__(self, header=None, body=None):
            self.header = header  # str
            self.body = body  # str

    doc_sections = OrderedDict([("", Section(header=""))])
    if not doc:
        return doc_sections

    doc = cleandoc(doc)
    lines = iter(doc.splitlines())

    header = ""
    body = []
    section = Section(header=header)
    line = ""
    while True:
        try:
            prev_line = line
            line = next(lines)
            # section header encountered
            if is_delimiter(line) and 0 < len(prev_line) <= len(line):
                # prev-prev-line is overline
                if (
                    len(body) >= 2
                    and len(body[-2]) == len(line)
                    and body[-2][0] == line[0]
                    and is_delimiter(body[-2])
                ):
                    lim = -2
                else:
                    lim = -1

                section.body = "\n".join(body[:lim]).rstrip()
                doc_sections.update([(header.strip(), section)])
                section = Section(header="\n".join(body[lim:] + [line]))
                header = prev_line
                body = []
                line = ""
            else:
                body.append(line)

        except StopIteration:
            section.body = "\n".join(body).rstrip()
            doc_sections.update([(header.strip(), section)])
            break
    return doc_sections


def merge_rest_docs(prnt_doc=None, child_doc=None):
    """ See custom_inherit.style_store.reST for details. """
    prnt_sections = parse_rest_doc(prnt_doc)
    child_sections = parse_rest_doc(child_doc)

    header = prnt_sections[""]
    prnt_sections.update(child_sections)
    if not child_sections[""].body:
        prnt_sections[""] = header
        if not header.body:
            prnt_sections.popitem(last=False)

    return "\n\n".join(
        ("\n".join((x.header, x.body)) for x in prnt_sections.values())
    ).lstrip()
